// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:http/http.dart';
import 'package:wyatt_http_client/src/authentication/interfaces/header_authentication_client.dart';
import 'package:wyatt_http_client/src/utils/digest_auth.dart';
import 'package:wyatt_http_client/src/utils/header_keys.dart';
import 'package:wyatt_http_client/src/utils/http_status.dart';
import 'package:wyatt_http_client/src/utils/utils.dart';

class DigestAuthenticationClient extends HeaderAuthenticationClient {
  final String username;
  final String password;
  final DigestAuth _digestAuth;
  final String authenticationHeader;
  final String wwwAuthenticateHeader;

  DigestAuthenticationClient({
    required this.username,
    required this.password,
    this.authenticationHeader = HeaderKeys.authorization,
    this.wwwAuthenticateHeader = HeaderKeys.wwwAuthenticate,
    BaseClient? inner,
  })  : _digestAuth = DigestAuth(username, password),
        super(inner);

  @override
  Map<String, String> modifyHeader(
    Map<String, String> header, [
    BaseRequest? request,
  ]) {
    if ((_digestAuth.isReady()) && request != null) {
      header[authenticationHeader] = _digestAuth.getAuthString(
        request.method,
        request.url,
      );
    }
    return header;
  }

  @override
  Future<StreamedResponse> send(BaseRequest request) async {
    // Check if our DigestAuth is ready.
    if (_digestAuth.isReady()) {
      // If it is, try to send the request with the modified header.
      return super.send(request);
    }

    // If it isn't, try to send the request without the modified header.
    final response = await inner.send(request);

    if (response.statusCode == HttpStatus.unauthorized) {
      final newRequest = Utils.copyRequest(request);
      final authInfo =
          response.headers[HeaderKeys.wwwAuthenticate.toLowerCase()];
      _digestAuth.initFromAuthenticateHeader(authInfo);
      return super.send(newRequest);
    } else {
      return response;
    }
  }
}
