// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:http/http.dart';
import 'package:wyatt_http_client/src/authentication/interfaces/oauth2_client.dart';
import 'package:wyatt_http_client/src/utils/authentication_methods.dart';
import 'package:wyatt_http_client/src/utils/header_keys.dart';
import 'package:wyatt_http_client/src/utils/http_status.dart';
import 'package:wyatt_http_client/src/utils/utils.dart';

class RefreshTokenClient extends Oauth2Client {
  final String authorizationEndpoint;
  final String tokenEndpoint;

  String? accessToken;
  final TokenParser accessTokenParser;
  String? refreshToken;
  final TokenParser refreshTokenParser;

  final String authenticationHeader;
  final String authenticationMethod;

  RefreshTokenClient({
    required this.authorizationEndpoint,
    required this.tokenEndpoint,
    required this.accessTokenParser,
    required this.refreshTokenParser,
    this.authenticationHeader = HeaderKeys.authorization,
    this.authenticationMethod = AuthenticationMethods.bearer,
    BaseClient? inner,
  }) : super(inner);

  @override
  Map<String, String> modifyHeader(
    Map<String, String> header, [
    BaseRequest? request,
  ]) {
    if (accessToken != null && request != null) {
      header[authenticationHeader] = '$authenticationMethod $accessToken';
      return header;
    }
    return header;
  }

  @override
  Future<void> authorize(
    Map<String, dynamic> body, {
    Map<String, String>? headers,
  }) async {
    final response = await inner.post(
      Uri.parse(authorizationEndpoint),
      body: jsonEncode(body),
      headers: headers,
    );

    if (response.statusCode == HttpStatus.ok) {
      final body = json.decode(response.body) as Map<String, dynamic>;
      final accessToken = accessTokenParser(body);
      final refreshToken = refreshTokenParser(body);

      if (accessToken.isNotEmpty) {
        this.accessToken = accessToken;
      }
      if (refreshToken.isNotEmpty) {
        this.refreshToken = refreshToken;
      }
    }
  }

  @override
  Future<void> refresh() async {
    if (refreshToken != null) {
      final Map<String, String> header = {
        authenticationHeader: '$authenticationHeader $refreshToken',
      };

      final response = await inner.get(
        Uri.parse(tokenEndpoint),
        headers: header,
      );

      if (response.statusCode == HttpStatus.ok) {
        final body = json.decode(response.body) as Map<String, dynamic>;
        accessToken = accessTokenParser(body);
      }
    }
  }

  @override
  Future<StreamedResponse> send(BaseRequest request) async {
    final response = await super.send(request);

    if (response.statusCode == HttpStatus.unauthorized) {
      await refresh();
      return super.send(Utils.copyRequest(request));
    }

    return response;
  }
}
